#!/usr/bin/env python
# --coding:utf-8--

# Copyright (c) 2020 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

import copy
import logging
from threading import RLock, Condition

from nebula2.common.ttypes import HostAddr, ErrorCode


class PartInfo(object):
    def __init__(self, part_id, leader: HostAddr, cursor=None):
        self.part_id = part_id
        self.leader = leader
        self.cursor = cursor
        self.has_done = False

    def __repr__(self) -> str:
        return 'PartInfo: part_id: {}, leader: {}, cursor: {}, has_done: {}'\
            .format(self.part_id, self.leader, self.cursor, self.has_done)


class PartManager(object):
    def __init__(self, parts):
        self._lock = RLock()
        self._condition = Condition()
        self._parts = parts
        if self._parts is None:
            self._parts = {}
        self._part_jobs = len(self._parts)
        self._stop = False

    def get_part(self, addr):
        try:
            self._condition.acquire()
            for part_id in self._parts.keys():
                if self._parts[part_id].leader == addr \
                        and not self._parts[part_id].has_done:
                    return self._parts[part_id]
            return None
        finally:
            self._condition.release()

    def update_part_info(self, part_info, is_finished):
        try:
            self._condition.acquire()
            if self._part_jobs > 0:
                self._part_jobs -= 1
            if part_info.part_id in self._parts.keys():
                if is_finished:
                    self._parts.pop(part_info.part_id)
                else:
                    self._parts[part_info.part_id] = part_info
            self._condition.notify_all()
        finally:
            self._condition.release()

    def update_part_leader(self, part_id, leader):
        try:
            self._condition.acquire()
            if part_id in self._parts.keys():
                self._parts[part_id].leader = leader
            self._condition.notify()
        finally:
            self._condition.release()

    def is_finish(self):
        try:
            self._condition.acquire()
            if self._part_jobs == 0 or self._stop:
                return True
            else:
                return False
        finally:
            self._condition.release()

    def set_stop(self):
        logging.debug("Stop the jobs")
        try:
            self._condition.acquire()
            self._stop = True
            self._condition.notify_all()
        finally:
            self._condition.release()

    def reset_jobs(self):
        logging.debug("Reset the jobs' status ")
        try:
            self._condition.acquire()
            if self._stop:
                return 
            self._part_jobs = len(self._parts)
            self._reset_parts_status()
        finally:
            self._condition.release()

    def _reset_parts_status(self):
        for part in self._parts:
            self._parts[part].has_done = False

    def has_next(self):
        try:
            self._condition.acquire()
            if self._stop:
                return False
            return len(self._parts) != 0
        finally:
            self._condition.release()

    def wait_task(self):
        try:
            self._condition.acquire()
            self._condition.wait()
        finally:
            self._condition.release()


def do_scan_job(storage_connection,
                parts_manager,
                in_req,
                scan_vertex=True,
                partial_success=False):
    data_sets = []
    req = copy.deepcopy(in_req)
    while True:
        is_finished = False  # the part without next, it means is finished
        if parts_manager.is_finish():
            break
        part_info = parts_manager.get_part(storage_connection.storage_addr())
        if part_info is None:
            parts_manager.wait_task()
            continue
        else:
            req.part_id = part_info.part_id
            logging.debug('Scan =====> req: {}'.format(req))
            if part_info.cursor is not None:
                req.cursor = part_info.cursor
            try:
                if scan_vertex:
                    resp = storage_connection.scan_vertex(req)
                else:
                    resp = storage_connection.scan_edge(req)
                logging.debug('Scan <==== get resp: {}'.format(resp))
                if len(resp.result.failed_parts) != 0:
                    if resp.result.failed_parts[0].code == ErrorCode.E_LEADER_CHANGED:
                        if resp.result.failed_parts[0].leader is None:
                            logging.error('Happen leader change, but the leader is None')
                            raise RuntimeError('Happen leader change, but the leader is None')
                        parts_manager.update_part_leader(resp.result.failed_parts[0].part_id,
                                                         resp.result.failed_parts[0].leader)
                        logging.warning('part_id {} has leader change, '
                                        'old leader is {}, new leader is {}'
                                        .format(part_info.part_id, storage_connection.storage_addr(),
                                                resp.result.failed_parts[0].leader))
                        storage_connection.update_leader_info(req.space_id,
                                                              req.part_id,
                                                              resp.result.failed_parts[0].leader)
                        continue
                    error = 'Query storage: {}, part id: {} failed: {}' \
                        .format(storage_connection.storage_addr(),
                                part_info.part_id, resp.result.failed_parts[0].code)
                    if not partial_success:
                        logging.error(error)
                        parts_manager.set_stop()
                        return error, []
                    logging.error(error)
                    is_finished = True
                    continue
                part_info.has_done = True
                if resp.has_next:
                    part_info.cursor = resp.next_cursor
                    logging.debug("Get next next_cursor: {}".format(resp.next_cursor))
                else:
                    is_finished = True
                if scan_vertex:
                    logging.debug("resp.vertex_data size: {}".format(len(resp.vertex_data.rows)))
                    if len(resp.vertex_data.column_names) == 0:
                        return 'Part id: {} return empty column names'.format(part_info.part_id)
                    if len(resp.vertex_data.rows) == 0:
                        continue
                    data_sets.append(resp.vertex_data)
                else:
                    logging.debug("resp.edge_data size: {}".format(len(resp.edge_data.rows)))
                    if len(resp.edge_data.column_names) == 0:
                        return 'Part id: {} return empty column names'.format(part_info.part_id)
                    if len(resp.edge_data.rows) == 0:
                        continue
                    data_sets.append(resp.edge_data)
            except Exception as e:
                import traceback
                logging.error(traceback.format_exc())
                parts_manager.set_stop()
                return str(e), None
            finally:
                parts_manager.update_part_info(part_info, is_finished)
    return None, data_sets



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.Thrift import TType

import logging
_log = logging.getLogger('nebula2.fbthrift.validator')

import sys
if sys.version_info[0] >= 3:
    basestring = str
    unicode = str
    long = int

class TValidator:
    tinfo = {
            # ttype: (type_name, python_type, min_value, max_value)
            TType.BOOL: ('BOOL', bool, None, None),
            TType.BYTE: ('BYTE', int, -128, 127),
            TType.DOUBLE: ('DOUBLE', float, None, None),
            TType.I16: ('I16', int, -32768, 32767),
            TType.I32: ('I32', int, -2147483648, 2147483647),
            TType.I64: ('I64', (int, long), None, None),
            TType.STRING: ('STRING', basestring, None, None),
            TType.UTF8: ('UTF8', unicode, None, None),
            }

    def __init__(self):
        self.custom_validators = {}

    def addClassValidator(self, name, validator):
        self.custom_validators[name] = validator

    def validate(self, msg):
        if not hasattr(msg, 'thrift_spec'):
            _log.error("Not a valid thrift object")
            return False

        name = msg.__class__.__name__
        return self.check_struct(name, msg, msg.thrift_spec)

    def check_basic(self, name, value, thrift_type):
        if thrift_type not in self.tinfo:
            _log.warn("%s Unrecognized thrift type %d. No validation done!",
                    name, thrift_type)
            return True

        t_name, python_type, v_min, v_max = self.tinfo[thrift_type]
        if not isinstance(value, python_type):
            error = "Value %s is not a %s" % (str(value), t_name)
        elif (v_min is not None and value < v_min) \
                or (v_max is not None and value > v_max):
            error = "Value %s not within %s boundaries" % (str(value), t_name)
        else:
            error = None

        if error is None:
            _log.debug("%s -> %s (type: %s) OK", name, str(value), t_name)
        else:
            _log.error("ERROR: %s is WRONG. %s", name, error)
        return error is None

    def check_map(self, name, value, k_type, k_specs, v_type, v_specs):
        _log.debug('%s - MAP check:', name)
        ok = True
        for k, v in value.items():
            if not self.check_type("%s key" % (name), k, k_type, k_specs):
                ok = False
            if not self.check_type("%s[%s]" % (name, k), v, v_type, v_specs):
                ok = False
        return ok

    def check_listset(self, name, value, v_type, v_specs):
        _log.debug('%s - LIST/SET check:', name)
        ok = True
        for i, v in enumerate(value):
            if not self.check_type("%s[%d]" % (name, i), v, v_type, v_specs):
                ok = False
        return ok

    def check_type(self, name, value, v_type, specs):
        if value is None:
            _log.debug("%s - NOT set", name)
            return True

        if v_type == TType.STRUCT:
            struct_specs = specs[1]
            ok = self.check_struct(name, value, struct_specs)
        elif v_type == TType.MAP:
            k_type = specs[0]
            k_specs = specs[1]
            v_type = specs[2]
            v_specs = specs[3]
            ok = self.check_map(name, value, k_type, k_specs, v_type, v_specs)
        elif v_type in (TType.LIST, TType.SET):
            v_type = specs[0]
            v_specs = specs[1]
            ok = self.check_listset(name, value, v_type, v_specs)
        else:
            ok = self.check_basic(name, value, v_type)

        return ok

    def check_struct(self, name, value, specs):
        _log.debug('%s - STRUCT check:', name)
        if specs is None:
            _log.error("%s - Empty thrift specs, can not be validated", name)
            return False

        ok = True
        for spec in specs:
            if spec is None:
                # Some fields of the struct might be not defined or
                # skipped old fields and their spec will be None
                continue
            f_name = name + "." + str(spec[2])
            f_type = spec[1]
            f_value = getattr(value, spec[2])
            f_specs = spec[3]
            if not self.check_type(f_name, f_value, f_type, f_specs):
                ok = False

        class_name = value.__class__.__name__
        if ok and class_name in self.custom_validators:
            if self.custom_validators[class_name](value):
                _log.debug('%s - Custom validator for class %s OK',
                    name, class_name)
            else:
                _log.error('%s - Custom validator for class %s failed!',
                        name, class_name)
                ok = False
        return ok



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

from nebula2.fbthrift.Thrift import TProcessor, TMessageType, TException
from nebula2.fbthrift.protocol import TProtocolDecorator, TMultiplexedProtocol

class TMultiplexedProcessor(TProcessor):
    def __init__(self):
        self.services = {}

    def registerProcessor(self, serviceName, processor):
        self.services[serviceName] = processor

    def setEventHandler(self, event_handler, serviceName=None):
        """ Set event handler for a service. If serviceName is None,
        set event handler for all services"""
        if serviceName is not None:
            if not serviceName in self.services:
                raise TException("Cannot set event handler for service " +
                        serviceName + ": no such service")
            else:
                self.services[serviceName].setEventHandler(event_handler)
        else:
            for processor in self.services.values():
                processor.setEventHandler(event_handler)

    def process(self, iprot, oprot, server_ctx=None):
        (name, type, seqid) = iprot.readMessageBegin()
        if type != TMessageType.CALL and type != TMessageType.ONEWAY:
            raise TException("TMultiplex protocol only supports CALL & ONEWAY")

        if sys.version_info[0] >= 3 and isinstance(name, bytes):
            name = name.decode('utf-8')
        index = name.find(TMultiplexedProtocol.SEPARATOR)
        if index < 0:
            raise TException("Service name not found in message name: " +
                name + ". Did you forget to use TMultiplexProtocol " +
                "in your client?")

        serviceName = name[0:index]
        call = name[index + len(TMultiplexedProtocol.SEPARATOR):]
        if sys.version_info[0] >= 3:
            call = call.encode('utf-8')
        if not serviceName in self.services:
            raise TException("Service name not found: " + serviceName +
                ". Did you forget to call registerProcessor()?")

        standardMessage = (
            call,
            type,
            seqid
        )
        return self.services[serviceName].process(StoredMessageProtocol(iprot,
            standardMessage), oprot, server_ctx)


class StoredMessageProtocol(TProtocolDecorator.TProtocolDecorator):
    def __init__(self, protocol, messageBegin):
        TProtocolDecorator.TProtocolDecorator.__init__(self, protocol)
        self.messageBegin = messageBegin

    def readMessageBegin(self):
        return self.messageBegin



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.protocol.TProtocol import TProtocolBase
from types import *

class TProtocolDecorator():
    def __init__(self, protocol):
        TProtocolBase(protocol)
        self.protocol = protocol

    def __getattr__(self, name):
        if hasattr(self.protocol, name):
            member = getattr(self.protocol, name)
        if type(member) in [MethodType, FunctionType, LambdaType,
                BuiltinFunctionType, BuiltinMethodType]:
            return lambda *args, **kwargs: self._wrap(member, args, kwargs)
        else:
            return member
        raise AttributeError(name)

    def _wrap(self, func, args, kwargs):
        if type(func) == MethodType:
            result = func(*args, **kwargs)
        else:
            result = func(self.protocol, *args, **kwargs)
        return result



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import traceback

from nebula2.fbthrift.server.TServer import TServerEventHandler
from nebula2.fbthrift.Thrift import (
    TException,
    TProcessor,
)
from nebula2.fbthrift.util.async_common import (
    AsyncioRpcConnectionContext,
    FramedProtocol,
    THeaderProtocol,
    ThriftHeaderClientProtocolBase,
    TReadWriteBuffer,
    WrappedTransport,
)


__all__ = [
    'ThriftAsyncServerFactory', 'ThriftClientProtocolFactory',
    'ThriftServerProtocolFactory',
]


logger = logging.getLogger(__name__)


#
# Thrift server support
#


async def ThriftAsyncServerFactory(
    processor, *, interface=None, port=0, loop=None, nthreads=None, sock=None,
    backlog=100, ssl=None, event_handler=None, protocol_factory=None
):
    """
    ThriftAsyncServerFactory(processor) -> asyncio.Server

    asyncio.Server factory for Thrift processors. In the spirit of "there
    should be one obvious way to do it", this server only supports the new
    THeader protocol.

    If `interface` is None (the default), listens on all interfaces. `port` is
    0 by default, which makes the OS allocate the port. Enumerate the returned
    server's "sockets" attribute to know the port in this case.

    If not given, the default event loop is used. If `nthreads` is given, the
    default executor for the event loop is set to a thread pool of up to
    `nthreads`.

    ssl is an instance of ssl.SSLContext. If None (default) or False SSL/TLS is
    not used.

    event_handler must be a subclass of nebula2.fbthrift.server.TServer. If None,
    nebula2.fbthrift.server.TServer.TServerEventHandler is used. Specify a custom handler
    for custom event handling (e.g. handling new connections)

    protocol_factory is a function that takes a triplet of
    (processor, event_handler, loop=None) and returns a `asyncio.Protocol` instance
    that will be passed to a call to `asyncio.create_server`. processor will be a
    subclass of `TProcessor`, event_handler will be a subclass of `TServer`, and
    loop is an `Optional[asyncio.AbstractEventLoop]`. If protocol_factory is None
    `ThriftHeaderServerProtocol` is used.

    Notes about the processor method handling:

    1. By default all methods are executed synchronously on the event loop.
       This can lead to poor performance if a single run takes long to process.

    2. Mark coroutines with `async def` if you wish to use `await`
       to call async services, schedule tasks with customized executors, etc.

    3. Mark methods with @run_on_thread if you wish to run them on the thread
       pool executor. Note that unless you're accessing C extensions which free
       the GIL, this is not going to win you any performance.

    Use this to initialize multiple servers asynchronously::

        loop = asyncio.get_event_loop()
        servers = [
            ThriftAsyncServerFactory(handler1, port=9090, loop=loop),
            ThriftAsyncServerFactory(handler2, port=9091, loop=loop),
        ]
        loop.run_until_complete(asyncio.wait(servers))
        try:
            loop.run_forever()   # Servers are initialized now
        finally:
            for server in servers:
                server.close()
    """

    if loop is None:
        loop = asyncio.get_event_loop()

    if not isinstance(processor, TProcessor):
        try:
            processor = processor._processor_type(processor, loop=loop)
        except AttributeError:
            raise TypeError(
                "Unsupported processor type: {}".format(type(processor)),
            )

    if nthreads:
        from concurrent.futures import ThreadPoolExecutor
        loop.set_default_executor(
            ThreadPoolExecutor(max_workers=nthreads),
        )
    ehandler = TServerEventHandler() if event_handler is None else event_handler
    protocol_factory = protocol_factory or ThriftHeaderServerProtocol
    pfactory = functools.partial(protocol_factory, processor, ehandler, loop)
    server = await loop.create_server(
        pfactory,
        interface,
        port,
        sock=sock,
        backlog=backlog,
        ssl=ssl,
    )

    if server.sockets:
        for socket in server.sockets:
            ehandler.preServe(socket.getsockname())

    return server


def ThriftServerProtocolFactory(processor, server_event_handler, loop=None):
    return functools.partial(
        ThriftHeaderServerProtocol, processor, server_event_handler, loop,
    )


class ThriftHeaderServerProtocol(FramedProtocol):

    def __init__(self, processor, server_event_handler, loop=None):
        super().__init__(loop=loop)
        self.processor = processor
        self.server_event_handler = server_event_handler
        self.server_context = None

    async def message_received(self, frame):
        # Note: we are using a single `prot` for in and out so that
        # we can support legacy clients that only understand FRAMED.
        # The discovery of what the client supports happens in iprot's
        # transport so we have to reuse a single one here.
        buf = TReadWriteBuffer(frame)
        prot = THeaderProtocol(buf)

        try:
            await self.processor.process(
                prot, prot, self.server_context,
            )
            msg = buf.getvalue()
            if len(msg) > 0:
                self.transport.write(msg)
        except TException as e:
            logger.warning("TException while processing request: %s", str(e))
            msg = buf.getvalue()
            if len(msg) > 0:
                self.transport.write(msg)
        except asyncio.CancelledError:
            self.transport.close()
        except BaseException as e:
            logger.error("Exception while processing request: %s", str(e))
            logger.error(traceback.format_exc())
            self.transport.close()

    def connection_made(self, transport):
        self.upgrade_transport(transport)

    def upgrade_transport(self, transport):
        self.transport = transport
        socket = self.transport.get_extra_info("socket")
        if socket is not None:
            self.server_context = AsyncioRpcConnectionContext(socket)
        self.server_event_handler.newConnection(self.server_context)

    def connection_lost(self, exc):
        self.server_event_handler.connectionDestroyed(self.server_context)


#
# Thrift client support
#


def ThriftClientProtocolFactory(
    client_class,
    loop=None,
    timeouts=None,
    client_type=None,
):
    return functools.partial(
        ThriftHeaderClientProtocol,
        client_class,
        loop,
        timeouts,
        client_type,
    )


class SenderTransport(WrappedTransport):
    async def _send(self):
        while True:
            msg = await self._queue.get()
            self._clean_producers()
            self._trans.write(msg)


class ThriftHeaderClientProtocol(ThriftHeaderClientProtocolBase):
    async def timeout_task(self, fname, seqid, delay):
        await asyncio.sleep(delay, loop=self.loop)
        self._handle_timeout(fname, seqid)

    def wrapAsyncioTransport(self, asyncio_transport):
        return SenderTransport(asyncio_transport, self, self.loop)
